@@ -1,7 +1,7 @@
 /* SPDX-License-Identifier: BSD-2-Clause */
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2020 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2023 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -198,17 +198,14 @@ struct dhcpcd_ctx {
 
 #ifdef PRIVSEP
 	struct passwd *ps_user;	/* struct passwd for privsep user */
-	pid_t ps_root_pid;
-	int ps_root_fd;		/* Privileged Actioneer commands */
-	int ps_log_fd;		/* chroot logging */
-	int ps_data_fd;		/* Data from root spawned processes */
-	struct eloop *ps_eloop;	/* eloop for polling root data */
 	struct ps_process_head ps_processes;	/* List of spawned processes */
-	pid_t ps_inet_pid;
-	int ps_inet_fd;		/* Network Proxy commands and data */
-	pid_t ps_control_pid;
-	int ps_control_fd;	/* Control Proxy - generic listener */
-	int ps_control_data_fd;	/* Control Proxy - data query */
+	struct ps_process *ps_root;
+	struct ps_process *ps_inet;
+	struct ps_process *ps_ctl;
+	int ps_data_fd;		/* data returned from processes */
+	int ps_log_fd;		/* chroot logging */
+	int ps_log_root_fd;	/* outside chroot log reader */
+	struct eloop *ps_eloop;	/* eloop for polling root data */
 	struct fd_list *ps_control;		/* Queue for the above */
 	struct fd_list *ps_control_client;	/* Queue for the above */
 #endif
@@ -270,6 +267,8 @@ int dhcpcd_ifafwaiting(const struct interface *);
 int dhcpcd_afwaiting(const struct dhcpcd_ctx *);
 void dhcpcd_daemonise(struct dhcpcd_ctx *);
 
+void dhcpcd_signal_cb(int, void *);
+
 void dhcpcd_linkoverflow(struct dhcpcd_ctx *);
 int dhcpcd_handleargs(struct dhcpcd_ctx *, struct fd_list *, int, char **);
 void dhcpcd_handlecarrier(struct interface *, int, unsigned int);
