@@ -301,7 +301,7 @@ ps_root_ioctldom(struct dhcpcd_ctx *ctx, uint16_t domain, unsigned long request,
     void *data, size_t len)
 {
 
-	if (ps_sendcmd(ctx, ctx->ps_root->psp_fd, domain,
+	if (ps_sendcmd(ctx, PS_ROOT_FD(ctx), domain,
 	    request, data, len) == -1)
 		return -1;
 	return ps_root_readerror(ctx, data, len);
@@ -327,7 +327,7 @@ ssize_t
 ps_root_route(struct dhcpcd_ctx *ctx, void *data, size_t len)
 {
 
-	if (ps_sendcmd(ctx, ctx->ps_root->psp_fd, PS_ROUTE, 0, data, len) == -1)
+	if (ps_sendcmd(ctx, PS_ROOT_FD(ctx), PS_ROUTE, 0, data, len) == -1)
 		return -1;
 	return ps_root_readerror(ctx, data, len);
 }
@@ -346,7 +346,7 @@ ps_root_indirectioctl(struct dhcpcd_ctx *ctx, unsigned long request,
 
 	strlcpy(buf, ifname, IFNAMSIZ);
 	memcpy(buf + IFNAMSIZ, data, len);
-	if (ps_sendcmd(ctx, ctx->ps_root->psp_fd, PS_IOCTLINDIRECT,
+	if (ps_sendcmd(ctx, PS_ROOT_FD(ctx), PS_IOCTLINDIRECT,
 	    request, buf, IFNAMSIZ + len) == -1)
 		return -1;
 	return ps_root_readerror(ctx, data, len);
@@ -358,7 +358,7 @@ ssize_t
 ps_root_ifignoregroup(struct dhcpcd_ctx *ctx, const char *ifname)
 {
 
-	if (ps_sendcmd(ctx, ctx->ps_root->psp_fd, PS_IFIGNOREGRP, 0,
+	if (ps_sendcmd(ctx, PS_ROOT_FD(ctx), PS_IFIGNOREGRP, 0,
 	    ifname, strlen(ifname) + 1) == -1)
 		return -1;
 	return ps_root_readerror(ctx, NULL, 0);
@@ -400,7 +400,7 @@ ps_root_sysctl(struct dhcpcd_ctx *ctx,
 		p += newlen;
 	}
 
-	if (ps_sendcmd(ctx, ctx->ps_root->psp_fd, PS_SYSCTL,
+	if (ps_sendcmd(ctx, PS_ROOT_FD(ctx), PS_SYSCTL,
 	    flags, buf, (size_t)(p - buf)) == -1)
 		return -1;
 
